#include "stdafx.h"

//   Gstreamer
#include "GstVideo.h"

#include "opencv2/highgui/highgui.hpp"
#include <gst/gst.h>
#include "gstappsink.h"

//            gstreamer
static void GstVideoOnError (GstBus *bus, GstMessage *msg, void *pThis)
{
	((CGstVideo*)pThis)->onError(msg);
}

//       gstreamer
static void GstVideoOnStateChanged (GstBus *bus, GstMessage *msg, void *pThis)
{
	((CGstVideo*)pThis)->onStateChanged(msg);
}

//   , ,  appsink  ,        gstreamer,   
static GstFlowReturn GstVideoOnNewBuffer (GstAppSink *elt, void *pThis)
{
	return ((CGstVideo*)pThis)->onNewBuffer();
}

//   , ,  appsink  ,     ,      
static GstFlowReturn GstVideoOnNewPreroll (GstAppSink *elt, void *pThis)
{
	return ((CGstVideo*)pThis)->onNewPreroll();
}

//   , ,     
static void GstVideoOnEOS (GstAppSink *elt, void *pThis)
{
	return ((CGstVideo*)pThis)->onEOS();
}

//       gstreamer.        gstreamer   
UINT GstVideoInit (LPVOID pParam)
{
	return ((CGstVideo*)pParam)->init();
}

CGstVideo::CGstVideo (const char *pipelineString_, const char *gstreamerPath_)
{
	pipeline = NULL;
	pGMainContext = NULL;
	pGMainLoop = NULL;
	pGstAppSink = NULL;
	pGstBuffer = NULL;
	pBufferGstCaps = NULL;

	//  pipeline  gstreamer
	pipelineString = g_strdup_printf("%s ! videoconvert ! appsink", pipelineString_);

	//   gstreamer
	gstreamerPath = g_strdup_printf("%s", gstreamerPath_);

	errorMessageString = NULL;
	infoMessageString = NULL;
}

CGstVideo::~CGstVideo ()
{
	//  
	release();

	g_free(pipelineString);
	g_free(gstreamerPath);

	cleartInfoMessage();
	cleartErrorMessage();
}

//   ?
bool CGstVideo::isPlaying ()
{
	GstState current, pending;
	GstClockTime timeout = 5 * GST_SECOND;

	if (!GST_IS_ELEMENT(pipeline))
		return false;

	if (!gst_element_get_state(GST_ELEMENT(pipeline),&current, &pending, timeout))
	{
		setErrorMessage("    ");
		return false;
	}

	return current == GST_STATE_PLAYING;
}

//   
void CGstVideo::start ()
{
	if (gst_element_set_state(GST_ELEMENT(pipeline), GST_STATE_PLAYING) == GST_STATE_CHANGE_FAILURE)
		setErrorMessage("    ");
}

//   
void CGstVideo::stop ()
{
	if (gst_element_set_state(GST_ELEMENT(pipeline), GST_STATE_NULL) == GST_STATE_CHANGE_FAILURE)
		setErrorMessage("    ");
}

//    
void CGstVideo::setErrorMessage (const gchar *message)
{
	cleartErrorMessage();
	errorMessageString = g_strdup_printf("%s", message);

	//  
	release();
}

//   
void CGstVideo::setInfoMessage (const gchar *message)
{
	cleartInfoMessage();
	infoMessageString = g_strdup_printf("%s", message);
}

//        gstreamer
void CGstVideo::onError (GstMessage *msg)
{
	GError *err;
	gchar *debug_info;
	gchar *message_string;

	gst_message_parse_error(msg, &err, &debug_info);
	message_string = g_strdup_printf(",    \"%s\": %s", GST_OBJECT_NAME(msg->src), err->message);
	g_clear_error(&err);
	g_free(debug_info);
	setErrorMessage(message_string);
	g_free(message_string);

	stop();
}

//       gstreamer
void CGstVideo::onStateChanged (GstMessage *msg)
{
	GstState old_state, new_state, pending_state;

	gst_message_parse_state_changed(msg, &old_state, &new_state, &pending_state);

	// Only pay attention to messages coming from the pipeline, not its children
	if (GST_MESSAGE_SRC(msg) == GST_OBJECT(pipeline))
	{
		gchar *message = g_strdup_printf("  %s", gst_element_state_get_name(new_state));
		setInfoMessage(message);
		g_free(message);
	}
}

// ,  appsink  ,        gstreamer,   
GstFlowReturn CGstVideo::onNewBuffer ()
{
	GstSample *pGstSample = NULL;
	GstMapInfo *pGstMapInfo = new GstMapInfo;

	// bail out if EOS
	if (gst_app_sink_is_eos(pGstAppSink))
		return GST_FLOW_ERROR;

	pGstSample = gst_app_sink_pull_sample(pGstAppSink);

	if (!pGstSample)
		return GST_FLOW_ERROR;

	pGstBuffer = gst_sample_get_buffer(pGstSample);

	if (!pGstBuffer)
	{
		gst_sample_unref(pGstSample);
		return GST_FLOW_ERROR;
	}

	// construct a frame header if we did not have any yet
	if (!pBufferGstCaps)
	{
		// reuse the caps ptr
		if (pBufferGstCaps)
			gst_caps_unref(pBufferGstCaps);

		pBufferGstCaps = gst_sample_get_caps(pGstSample);

		// bail out in no caps
		assert(gst_caps_get_size(pBufferGstCaps) == 1);
	}

	// gstreamer expects us to handle the memory at this point so we can just wrap the raw buffer and be done with it the data ptr in GstMapInfo is only valid throughout the mapifo objects life.
	// TODO: check if reusing the mapinfo object is ok.
	if (!gst_buffer_map(pGstBuffer, pGstMapInfo, GST_MAP_READ))
	{
		gst_sample_unref(pGstSample);
		return GST_FLOW_ERROR;
	}

	//     
	pFrameBuffer = (char*)pGstMapInfo->data;

	//  
	gst_buffer_unmap(pGstBuffer, pGstMapInfo);
	gst_sample_unref(pGstSample);
	delete pGstMapInfo;

	return GST_FLOW_OK;
}

// ,  appsink  ,     ,      
GstFlowReturn CGstVideo::onNewPreroll ()
{
	setInfoMessage("GstVideoOnNewPreroll");

	return GST_FLOW_OK;
}

// ,     
void CGstVideo::onEOS ()
{
	setInfoMessage("GstVideoOnEOS");
}

//       gstreamer.        gstreamer   
UINT CGstVideo::init ()
{
	GError *error = NULL;

	//  gstreamer
	gst_init(NULL, NULL);

	//    gstreamer   ,      
	GstRegistry *pGstRegistry = gst_registry_get();
	if (!gst_registry_scan_path(pGstRegistry, gstreamerPath))
	{
		gchar *message = g_strdup_printf("    gstreamer: %s", gstreamerPath);
		setErrorMessage(message);
		g_free(message);

		return 0;
	}

	// Create our own GLib Main Context and make it the default one
	pGMainContext = g_main_context_new();
	g_main_context_push_thread_default(pGMainContext);

	// Build pipeline
	pipeline = gst_parse_launch(pipelineString, &error);
	if (error)
	{
		gchar *message = g_strdup_printf("   : %s", error->message);
		g_clear_error(&error);
		setErrorMessage(message);
		g_free (message);

		return 0;
	}

	// Set the pipeline to READY, so it can already accept a window handle, if we have one
	start();

	GstIterator *it = gst_bin_iterate_sinks(GST_BIN(pipeline));
	gboolean done = FALSE;
	GstElement *element = NULL;
	gchar *name = NULL;
	GValue value = G_VALUE_INIT;

	while (!done)
	{
		switch (gst_iterator_next(it, &value))
		{
			case GST_ITERATOR_OK:
				element = GST_ELEMENT(g_value_get_object(&value));
				name = gst_element_get_name(element);

				if (name)
				{
					if (strstr(name, "autovideosink") != NULL || strstr(name, "appsink") != NULL)
					{
						pGstAppSink = (GstAppSink *)gst_object_ref(element);
						done = TRUE;
					}

					g_free(name);
				}
				g_value_unset(&value);

				break;
			case GST_ITERATOR_RESYNC:
				gst_iterator_resync(it);
				break;
			case GST_ITERATOR_ERROR:
			case GST_ITERATOR_DONE:
				done = TRUE;
				break;
		}
	}
	gst_iterator_free(it);

	if (!pGstAppSink)
	{
		setErrorMessage("   ");

		return 0;
	}

	// Instruct the bus to emit signals for each received message, and connect to the interesting signals
	GstBus *bus = gst_element_get_bus(pipeline);
	GSource *bus_source = gst_bus_create_watch(bus);
	g_source_set_callback(bus_source, (GSourceFunc)gst_bus_async_signal_func, NULL, NULL);
	g_source_attach(bus_source, pGMainContext);
	g_source_unref(bus_source);
	g_signal_connect(G_OBJECT(bus), "message::error", (GCallback)GstVideoOnError, this);
	g_signal_connect(G_OBJECT(bus), "message::state-changed", (GCallback)GstVideoOnStateChanged, this);
	gst_object_unref(bus);

	// TODO: is 1 single buffer really high enough?
	gst_app_sink_set_max_buffers(pGstAppSink, 1);
	gst_app_sink_set_drop(pGstAppSink, true);
	// do not emit signals: all calls will be synchronous and blocking
	gst_app_sink_set_emit_signals(pGstAppSink, false);

	// support 1 and 3 channel 8 bit data, as well as bayer (also 1 channel, 8bit)
	GstCaps *caps = gst_caps_from_string("video/x-raw, format=(string){BGR, GRAY8}; video/x-bayer,format=(string){rggb, bggr, grbg, gbrg}");
	gst_app_sink_set_caps(pGstAppSink, caps);
	gst_caps_unref(caps);

	//   
	GstAppSinkCallbacks gstCallbacks = {NULL};
	gstCallbacks.eos = &GstVideoOnEOS;
	gstCallbacks.new_preroll = &GstVideoOnNewPreroll;
	gstCallbacks.new_sample = &GstVideoOnNewBuffer;

	gst_app_sink_set_callbacks(pGstAppSink, &gstCallbacks, this, NULL);

	// Create a GLib Main Loop and set it to run
	pGMainLoop = g_main_loop_new(pGMainContext, FALSE);
	g_main_loop_run(pGMainLoop);

	//  
	release();

	return 0;
}

//  
void CGstVideo::release ()
{
	if (isPlaying())
		stop();

	if (pGMainLoop)
	{
		g_main_loop_unref(pGMainLoop);
		pGMainLoop = NULL;
	}

	if (pGMainContext)
	{
		g_main_context_pop_thread_default(pGMainContext);
		g_main_context_unref(pGMainContext);
	}

	if (pipeline)
	{
		stop();
		gst_object_unref(pipeline);
	}

	if (pGstAppSink)
		gst_object_unref(pGstAppSink);

	if (pGstBuffer)
		gst_buffer_unref(pGstBuffer);

	if (pFrameBuffer)
		pFrameBuffer = NULL;

	if (pBufferGstCaps)
		gst_caps_unref(pBufferGstCaps);
}

//      
gint64 CGstVideo::getPositionByMsec ()
{
	gint64 value;

	if (!pipeline || !pGstAppSink)
	{
		setInfoMessage("  ");
		return 0;
	}

	if (!gst_element_query_position((GstElement*)pGstAppSink, GST_FORMAT_TIME, &value))
	{
		setInfoMessage("     ");
		return 0;
	}
	
	return (gint64)(value * 1e-6); // nano seconds to milli seconds
}

//      
gint64 CGstVideo::getPositionByFrames ()
{
	gint64 value;

	if (!pipeline || !pGstAppSink)
	{
		setInfoMessage("  ");
		return 0;
	}

	if (!gst_element_query_position((GstElement*)pGstAppSink, GST_FORMAT_DEFAULT, &value))
	{
		setInfoMessage("     ");
		return 0;
	}

	return value;
}

//      
gint64 CGstVideo::getPositionByPercent ()
{
	gint64 value;

	if (!pipeline || !pGstAppSink)
	{
		setInfoMessage("  ");
		return 0;
	}

	if (!gst_element_query_position((GstElement*)pGstAppSink, GST_FORMAT_PERCENT, &value))
	{
		setInfoMessage("     ");
		return 0;
	}

	return (gint64)(((double) value) / GST_FORMAT_PERCENT_MAX);
}

//    
gint CGstVideo::getWidth ()
{
	if (!pipeline || !pGstAppSink)
	{
		setInfoMessage("  ");
		return 0;
	}

	if (!pBufferGstCaps)
	{
		setInfoMessage("     ,    ");
		return 0;
	}

	GstStructure* structure = gst_caps_get_structure(pBufferGstCaps, 0);
	gint width = 0;

	if (!gst_structure_get_int(structure, "width", &width))
	{
		setInfoMessage("     ");
		return 0;
	}

	return width;
}

//    
gint CGstVideo::getHeight ()
{
	if (!pipeline || !pGstAppSink)
	{
		setInfoMessage("  ");
		return 0;
	}

	if (!pBufferGstCaps)
	{
		setInfoMessage("     ,    ");
		return 0;
	}

	GstStructure* structure = gst_caps_get_structure(pBufferGstCaps, 0);
	gint height = 0;

	if (!gst_structure_get_int(structure, "height", &height))
	{
		setInfoMessage("     ");
		return 0;
	}

	return height;
}

//  FPS 
gint CGstVideo::getFPS ()
{
	if (!pipeline || !pGstAppSink)
	{
		setInfoMessage("  ");
		return 0;
	}

	if (!pBufferGstCaps)
	{
		setInfoMessage("   FPS ,    ");
		return 0;
	}

	GstStructure* structure = gst_caps_get_structure(pBufferGstCaps, 0);
	gint num = 0, denom = 1;
	if (!gst_structure_get_fraction(structure, "framerate", &num, &denom))
	{
		setInfoMessage("   FPS ");
		return 0;
	}

	return (gint)(num/denom);
}

//     
gint CGstVideo::getQueueLenght ()
{
	if (!pipeline || !pGstAppSink)
	{
		setInfoMessage("  ");
		return 0;
	}

	return gst_app_sink_get_max_buffers(pGstAppSink);
}

//     
gint CGstVideo::getChannels ()
{
	if (!pipeline || !pGstAppSink)
	{
		setInfoMessage("  ");
		return 0;
	}

	if (!pBufferGstCaps)
	{
		setInfoMessage("      ,    ");
		return 0;
	}

	GstStructure* structure = gst_caps_get_structure(pBufferGstCaps, 0);
	int channels = 1;
	const gchar* name = gst_structure_get_name(structure);
	const gchar* format = gst_structure_get_string(structure, "format");

	if (!name || !format)
	{
		setInfoMessage("      ");
		return 0;
	}

	// we support 3 types of data:
	//     video/x-raw, format=BGR   -> 8bit, 3 channels
	//     video/x-raw, format=GRAY8 -> 8bit, 1 channel
	//     video/x-bayer             -> 8bit, 1 channel
	// bayer data is never decoded, the user is responsible for that
	// everything is 8 bit, so we just test the caps for bit depth
	if (strcasecmp(name, "video/x-raw") == 0)
	{
		if (strcasecmp(format, "BGR") == 0)
			channels = 3;
		else if (strcasecmp(format, "GRAY8") == 0)
			channels = 1;
	}
	else if (strcasecmp(name, "video/x-bayer") == 0)
		channels = 1;

	return channels;
}

//     
char *CGstVideo::getFrameBuffer ()
{
	return pFrameBuffer;
}

//     
gchar *CGstVideo::getErrorMessage ()
{
	return errorMessageString;
}

//      
void CGstVideo::cleartErrorMessage ()
{
	if (errorMessageString)
		g_free(errorMessageString);

	errorMessageString = NULL;
}

//    
gchar *CGstVideo::getInfoMessage ()
{
	return infoMessageString;
}

//    
void CGstVideo::cleartInfoMessage ()
{
	if (infoMessageString)
		g_free(infoMessageString);

	infoMessageString = NULL;
}

//         
int CGstVideo::strcasecmp (const char *s1, const char *s2)
{
	while (1)
	{
		int c1 = tolower((unsigned char) *s1++);
		int c2 = tolower((unsigned char) *s2++);

		if (c1 == 0 || c1 != c2)
			return c1 - c2;
	}
}
